<?php

namespace Drupal\commerce_payrexx_integration;

use Drupal\commerce_order\Entity\Order;
use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payrexx_integration\Event\PaymentResponseEvent;
use Drupal\commerce_payrexx_integration\Event\PayrexxEvents;
use Drupal\commerce_price\Price;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Payrexx\Models\Response\Transaction;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Base service for handling Payrexx transactions.
 */
class TransactionHandlerService {

  const SUCCESS_RESPONSES = [
    Transaction::CONFIRMED,
  ];

  const PARTIAL_SUCCESS_RESPONSES = [
    Transaction::INITIATED,
    Transaction::WAITING,
    Transaction::AUTHORIZED,
    Transaction::RESERVED,
  ];

  const FAILED_RESPONSES = [
    Transaction::DECLINED,
    Transaction::ERROR,
    Transaction::EXPIRED,
    Transaction::INSECURE,
    Transaction::CANCELLED,
    Transaction::UNCAPTURED,
  ];

  const REFUND_RESPONSES = [
    Transaction::PARTIALLY_REFUNDED,
    Transaction::REFUND_PENDING,
    Transaction::REFUNDED,
  ];

  /**
   * The Payrexx redirect checkout payment gateway plugin.
   *
   * @var \Drupal\commerce_payrexx_integration\Plugin\Commerce\PaymentGateway\PayrexxRedirectCheckout
   */
  protected $redirectCheckout;

  /**
   * The EntityTypeManager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The logger channel for this module.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;
  /**
   * TransactionHandlerService constructor.
   *
   * @param \Drupal\Core\Session\AccountInterface $currentUser
   */

  /**
   * Constructor for TransactionHandlerService.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, EventDispatcherInterface $eventDispatcher, LoggerChannelInterface $loggerChannel) {
    $this->entityTypeManager = $entityTypeManager;
    $this->eventDispatcher = $eventDispatcher;
    $this->logger = $loggerChannel;
  }

  /**
   * Success response handler.
   */
  protected function handleResponseSuccess(OrderInterface $order, Transaction $transaction) {
    $this->dispatchEvent($order, $transaction);
    $this->saveCommercePayment($order, $transaction, 'completed');
  }

  /**
   * Partial success response handler.
   */
  protected function handleResponsePartialSuccess(OrderInterface $order, Transaction $transaction) {
    $this->logger->info('Received a partially successful payment response for order %order: %details', [
      '%order' => $order->id(),
      '%details' => json_encode((array) $transaction),
    ]);

    $this->dispatchEvent($order, $transaction);
    $this->saveCommercePayment($order, $transaction, 'completed');
  }

  /**
   * Error response handler.
   */
  protected function handleResponseError(OrderInterface $order, Transaction $transaction) {
    $this->logger->warning('Received an error payment response for order %order: %details', [
      '%order' => $order->id(),
      '%details' => json_encode((array) $transaction),
    ]);

    $this->dispatchEvent($order, $transaction);

    throw new PaymentGatewayException('Payment incomplete or declined');
  }

  /**
   * Retrieve reference id and convert it to order id.
   */
  public function getOrderIdFromReferenceId(string $reference_id) {
    return (int) $reference_id;
  }

  /**
   * Create payment entity and attach it to the order.
   */
  protected function saveCommercePayment(OrderInterface $order, $response, $state) {
    /** @var \Drupal\commerce_payment\PaymentStorage $paymentStorage */
    $paymentStorage = $this->entityTypeManager->getStorage('commerce_payment');
    $payment = $paymentStorage->loadByRemoteId($response->getId());

    if (!$payment) {
      $payment = $paymentStorage->create(
        [
          'payment_gateway' => $this->redirectCheckout ? $this->redirectCheckout->getEntityId() : $order->get('payment_gateway')->target_id,
          'order_id' => $order->id(),
          'remote_id' => $response->getId(),
        ]
      );
    }
    $invoice = $response->getInvoice();
    $type = gettype($invoice);

    $amount = new Price($response->getAmount() / 100, $type === 'array' ? $invoice['currencyAlpha3'] : $invoice->currency);

    $payment->set('state', $state);
    $payment->set('amount', $amount);
    $payment->set('remote_state', $response->getStatus());
    $payment->save();
  }

  /**
   * Handle the response based on the response status.
   *
   * @param \Drupal\commerce_order\Entity\Order $order
   *   A commerce order.
   * @param \Payrexx\Models\Response\Transaction $transaction
   *   The post-payment transaction.
   */
  public function responseHandler(Order $order, Transaction $transaction) {
    if (in_array($transaction->getStatus(), TransactionHandlerService::SUCCESS_RESPONSES)) {
      $this->handleResponseSuccess($order, $transaction);
    }
    elseif (in_array($transaction->getStatus(), TransactionHandlerService::PARTIAL_SUCCESS_RESPONSES)) {
      $this->handleResponsePartialSuccess($order, $transaction);
    }
    elseif (in_array($transaction->getStatus(), TransactionHandlerService::FAILED_RESPONSES)) {
      $this->handleResponseError($order, $transaction);
    }
    elseif (in_array($transaction->getStatus(), TransactionHandlerService::REFUND_RESPONSES)) {
      $this->logger->warning('Received a currently unhandled payment refund response for order %order: %details', [
        '%order' => $order->id(),
        '%details' => json_encode((array) $transaction),
      ]);
    }
    else {
      $this->logger->warning('Received a currently unhandled payment response for order %order: %details', [
        '%order' => $order->id(),
        '%details' => json_encode((array) $transaction),
      ]);
    }
  }

  /**
   * Dispatch the PaymentResponseEvent.
   *
   * @param \Drupal\commerce_order\Entity\OrderInterface $order
   *   A commerce order.
   * @param \Payrexx\Models\Response\Transaction $transaction
   *   The post-payment transaction.
   */
  protected function dispatchEvent(OrderInterface $order, Transaction $transaction) {
    $this->eventDispatcher->dispatch(PayrexxEvents::PAYMENT_RESPONSE, new PaymentResponseEvent($order, $transaction));
  }

}
